function Y = OLS123(csvfilename)
%Sensor calibration by non weighted polynomial functions of order d = 1 to 3
%JLBK 31 March 2007, 15 Jan. 2008, 12 Feb. 2008, 18 Aug. 2020
% 
%This code:
%- runs an OLS regression for three polynomial functions written
%  y = b0 + b1*x, y = b0 + b1*x + b2*x^2 and y = b0 + b1*x + b2*x^2 + b3*x^3
%  to calculate the bj coefficients
%- calculates standard uncertainties and covariance of bj coefficients
%- compares the three functions to choose the optimal order d
%  by means of a Snedecor test
%- provides some graphs in addition to the numerical outputs
%
%Input: 
%- csv file name, without the '.csv' extension, with separator ;,
%  with one single line of header
%  data organised in columns in the following order:
%  xi, u(xi), and then Ny columns for the yik values.
%  All columns must contain the same number Nx of measured values.
%
%Note:
% the values u(xi) are not used by OLS123 but are used by WLS123Cal.
%
%Outputs:
% results are given in a file named OLS123.dat which contains successively
%- the optimal order dopt
%  and then for each function:
%- order
%- values of bj coefficients
%- standard unertainties of bj coefficients
%- covariances of bj coefficients
%- residual variance sl^2.
%
% options = optimset('TolFun',1e-15,'TolX',1e-15, 'MaxFunEvals',5000, 'MaxIter', 5000);

% reading data from csvfilename by ignoring its first line
D = dlmread([csvfilename,'.csv'],';',1,0);
Nx = size(D,1);
Ny = size(D,2)-2;
N = Nx*Ny;

% standard values xi
xi = D(:,1);
% range of xi values for the graphs
rangexi = (0: 1.1*max(xi)/100: 1.1*max(xi))';

% creation of the vector Y containing all yik values
Y = (reshape(D(:,3:Ny+2)',[],1));
% and of the matrix F containing ones in the 1st column
% and standard values xi in increasing order in the 2nd column
F(1:N,1)=ones;
F(:,2) = reshape(repmat(xi,1,Ny)',[],1);

% 1st order regression
d = 1;
% QR decomposition of F
[Q,R]=qr(F,0);
M=R\Q';
% calculation of bj
B1=M*Y;
% calculation of residuals
e = F*B1 - Y;
% calculation of residual variance sr and of bj covariance
cmr1 = e'*e;
sr1 = cmr1/(N-d-1);
varb1 = sr1*(M*M');
out1 = polyval(fliplr(B1'),rangexi);
% graph
titre1(1) = {'OLS 1st order polynomial calibration function'};
titre1(2) = cellstr(['y = ' num2str(B1(1)) ' + ' num2str(B1(2)) ' x']);
createfigureCalib(F(:,2),Y, rangexi, out1, csvfilename, titre1);

% 2nd order regression
d = 2;
% addition of the column of values xi^2 to matrix F 
F(:,3) = F(:,2).^2;
% QR decomposition of F
[Q,R]=qr(F,0);
M=R\Q';
% calculation of bj
B2=M*Y;
% calculation of residuals
e = F*B2 - Y;
% calculation of residual variance sr and of bj covariance
cmr2 = e'*e;
sr2 = cmr2/(N-d-1);
varb2 = sr2*(M*M');
out2 = polyval(fliplr(B2'),rangexi);
% graph
titre2(1) = {'OLS 2nd order polynomial calibration function'};
titre2(2) = cellstr(['y = ' num2str(B2(1)) ' + ' num2str(B2(2)) ' x + '...
    num2str(B2(3)) ' x^2']);
createfigureCalib(F(:,2),Y, rangexi, out2, csvfilename, titre2);

% 3rd order regression
d = 3;
% addition of the column of values xi^3 to matrix F 
F(:,4) = F(:,2).^3;
% QR decomposition of F
[Q,R]=qr(F,0);
M=R\Q';
% calculation of bj
B3=M*Y;
% calculation of residuals
e = F*B3 - Y;
% calculation of residual variance sr and of bj covariance
cmr3 = e'*e;
sr3 = cmr3/(N-d-1);
varb3 = sr3*(M*M');
out3 = polyval(fliplr(B3'),rangexi);

% graph
titre3(1) = {'OLS 3rd order polynomial calibration function'};
titre3(2) = cellstr(['y = ' num2str(B3(1)) ' + ' num2str(B3(2)) ' x + '...
    num2str(B3(3)) ' x^2 + ' num2str(B3(4)) 'x^3']);
createfigureCalib(F(:,2),Y, rangexi, out3, csvfilename, titre3);

% graph of the variance of the yik values
titre(1) = {['data ',csvfilename,'.csv']};
titre(2) = {'variance of the yik values'};
createfigureVar(xi,((std(D(:,3:Ny+2)'))'.^2)',csvfilename,titre);

% Snedecor test and determination of the optimal order dopt
F1calc = (N-3)*(cmr1-cmr2)/cmr2;
F2calc = (N-4)*(cmr2-cmr3)/cmr3;
F1theo = finv(0.95,1,N-3);
F2theo = finv(0.95,1,N-4);
if F1calc < F1theo
    degopt = 1;
else
    if F2calc < F2theo
        degopt = 2;
    else, degopt = 3;
    end
end

% final results
Y = [degopt;
    1; B1(1); B1(2); sqrt(varb1(1,1)); sqrt(varb1(2,2)); varb1(1,2); sr1;...
    2; B2(1); B2(2); B2(3); sqrt(varb2(1,1)); sqrt(varb2(2,2)); sqrt(varb2(3,3)); varb2(1,2); varb2(1,3); varb2(2,3); sr2;...
    3; B3(1); B3(2); B3(3); B3(4); sqrt(varb3(1,1)); sqrt(varb3(2,2)); sqrt(varb3(3,3)); sqrt(varb3(4,4));...
        varb3(1,2); varb3(1,3); varb3(1,4); varb3(2,3); varb3(2,4); varb3(3,4); sr3];
% save results in OLS123.mat variable
save('OLS123.mat','Y');

function createfigureCalib(x1, y1, x2, y2, csvfilename, Title)
%JLBK Auto-generated by MATLAB on 23-Mar-2007 08:38:51

%% Create figure
figure1 = figure;
 
%% Create axes
axes1 = axes(...
  'XGrid','on',...
  'XMinorTick','on',...
  'YGrid','on',...
  'YMinorTick','on',...
  'Parent',figure1);
axis(axes1,[0 1.1*max(x1) 0 1.1*max(y1)]);
title(axes1,Title);
xlabel(axes1,'quantity X');
ylabel(axes1,'quantity Y');
box(axes1,'on');
hold(axes1,'all');
 
%% Create plot
plot1 = plot(...
  x1,y1,...
  'LineStyle','none',...
  'Color','k',...
  'Marker','*',...
  'Parent',axes1);
 
%% Create mutliple lines using matrix input to plot
plot2 = plot(x2,y2);
set(plot2,...
  'Color',[0.4941 0.4941 0.4941],...
  'LineStyle','-',...
  'Color','b',...
  'LineWidth',1);
 
%% Create legend
legend1 = legend(axes1,{['data ',csvfilename,'.csv'],'calibration'},'Location','Best');

function createfigureVar(x1, y1, csvfilename, Title)
%JLBK Auto-generated by MATLAB on 23-Mar-2007 08:38:51

%% Create figure
figure1 = figure;
 
%% Create axes
axes1 = axes(...
  'XGrid','on',...
  'XMinorTick','on',...
  'YGrid','on',...
  'YMinorTick','on',...
  'Parent',figure1);
axis(axes1,[0 1.1*max(x1) 0 1.1*max(y1)]);
title(axes1,Title);
xlabel(axes1,'calibration range');
ylabel(axes1,'variance si ^2');
box(axes1,'on');
hold(axes1,'all');
 
%% Create plot
plot1 = plot(...
  x1,y1,...
  'LineStyle',':',...
  'Color','b',...
  'Marker','o',...
  'Parent',axes1);

%% Create legend
legend1 = legend(axes1,{'initial variance'},'Location','Best');