function Y = WLS123cal(csvfilename,NMC,PreRel,silim)
% Williamson least squares regression for polynomial functions
% of order d = 1 to 3
% adapted to sensor calibration functions with Monte Carlo simulations.
% JLBK, 23 March 2007, 15 Jan. 2008, 10 March 2013, 23 Aug. 2020

% This code runs Williamson regression for polynomial functions written
% y = b0 + b1*x + b2*x^2 + b3*x^3
% It calculates
% - the bj coefficients by accounting for uncertainties in both xi and yik values.
% - standard uncertainties and co-variances of bj coefficients.
% - replacement of yik variances if the are initially equal to zero
%
% Inputs: 
%- csv file name, without the '.csv' extension, with separator ;,
%  with one single line of header
%  data organised in columns in the following order:
%  xi, u(xi), and then Ny columns for the yik values.
%  All columns must contain the same number Nx of measured values.
%- NMC: number of Monte Carlo simulations Monte-Carlo, with NMC >= 500.
%- PreRel: relative accuracy on final bj coefficients values, to be given as 1e-k.
%- silim: minimal significant variance. Default value can be 1e-6. To be
%  adapted by the user.
%
%Outputs:
% results are given in a file named OLS123.dat which contains successively
%- the optimal order dopt
%  and then for each function:
%- order
%- values of bj coefficients
%- standard unertainties of bj coefficients
%- covariances of bj coefficients
%- residual variance sl^2.


% global variables for functions Willorderd(z)
global Nx Ny deg1 deg2 deg3 xi wxi all_yik all_wyi;

% set of controls
options = optimset('TolFun',1e-10,'TolX',1e-10, 'MaxFunEvals',50000, 'MaxIter', 50000);
% definition
name_calibration = 'Etalonnage / calibration';

% polynomial orders
deg1 = 1;
deg2 = 2;
deg3 = 3;

% reading data from csvfilename by ignoring its first line
D = dlmread([csvfilename,'.csv'],';',1,0);
Nx = size(D,1);
Ny = size(D,2)-2;
N = Nx*Ny;

% standard values xi
xi = D(:,1);
% column vector with all ordered xi * Ny
all_xi = reshape(repmat(xi,1,Ny)',[],1);
% standard uncertainties u(xi)
uxi = D(:,2);
% weights Wxi
wxi = 1./uxi.^2;

% creation of column the vector of all yik values from the csv file matrix
all_yik = (reshape(D(:,3:Ny+2)',[],1));
% initialisation
all_yi_init = all_yik;
% standard unecrtainties u(yi)
% calculation of standard deviations si
si = std(D(:,3:(Ny+2))');
% in case si < silim (i.e. close or equal to zero),
% then remplace si by the minimum non zero value
si = min(si + 1e10*(si < silim)).*(si < silim) + si.*(si >= silim);
% weights Wyi
wyi = 1./si.^2;
% column vector of all ordered weights Wyi
all_wyi = reshape(repmat(wyi',1,Ny)',[],1);

% range of xi values for the final calculations of derivatives and graphs
rangexi = (0: 1.1*max(xi)/100: 1.1*max(xi))';

% For the regress function, creation of a matrix A with 3 columns
% containing
% in 1st column: yik values,
% in 2nd column: ones
% in 3rd column: xi values in increasing order
A(:,1)=all_yik;
A(:,2)=ones;
A(:,3)=all_xi;

% d=1 OLS regression
B1=regress(A(:,1),A(:,2:3), 0.05);
bj01 = B1';
bj01_init = bj01;

% d=2 OLS regression
% add to A the column of xi^2 values
A(:,4)=A(:,3).^2;
B2=regress(A(:,1),A(:,2:4), 0.05);
bj02 = B2';
bj02_init = bj02;

% d=3 OLS regression
% add to A the column of xi^3 values
A(:,5)=A(:,3).^3;
B3=regress(A(:,1),A(:,2:5), 0.05);
bj03 = B3';
bj03_init = bj03;

% Initialisation of samples for MC simulations
xiMC=zeros(Nx,NMC);
yiMC=zeros(Nx*Ny,NMC);

rng('default');
% creation of xi samples
for i=1:Nx
    xiMC(i,:) = randn(1,NMC)*sqrt(1./wxi(i)) + xi(i);
end
% creation of yik samples
for i=1:Nx*Ny
    yiMC(i,:) = randn(1,NMC)*sqrt(1./all_wyi(i)) + A(i,1);
end

% declaration of result variables for d = 1:3
Mbj1=zeros(NMC,deg1+1);
Mbj2=zeros(NMC,deg2+1);
Mbj3=zeros(NMC,deg3+1);

% calculation of the bj coefficients for the NMC samples
for i=1:NMC
    xi = xiMC(:,i);
    all_yi = yiMC(:,i);
    % initial values of Xi = xi
    Z01 = vertcat(B1,xi);
    Z02 = vertcat(B2,xi);
    Z03 = vertcat(B3,xi);
    % initial control variable
    control = 1;
    % order 1
    while sum(abs(control) < PreRel) <= (Nx+deg1)
        [Z11, fval1] = fminsearch(@Willorder1,Z01,options);
        % replace the values equal to zero in Z0 in the denominator
        % by 1e-10 to avoid divisions by zero
        control = (Z11-Z01)./(Z01 + 1e-10.*(Z01==0));
        Z01 = Z11;
    end %end of the loop while sum(abs(control1) < PreRel) <= (Nx+deg1)
    Mbj1(i,:) = Z01(1:deg1+1)';
    % order 2
    while sum(abs(control) < PreRel) <= (Nx+deg2)
        [Z12, fval2] = fminsearch(@Willorder2,Z02,options);
        % replace the values equal to zero in Z0 in the denominator
        % by 1e-10 to avoid divisions by zero
        control = (Z12-Z02)./(Z02 + 1e-10.*(Z02==0));
        Z02 = Z12;
    end %end of the loop while sum(abs(control2) < PreRel) <= (Nx+deg2)
    Mbj2(i,:) = Z02(1:deg2+1)';
    % order 3
    while sum(abs(control) < PreRel) <= (Nx+deg3)
        [Z13, fval3] = fminsearch(@Willorder3,Z03,options);
        % replace the values equal to zero in Z0 in the denominator
        % by 1e-10 to avoid divisions by zero
        control = (Z13-Z03)./(Z03 + 1e-10.*(Z03==0));
        Z03 = Z13;
    end %end of the loop while sum(abs(control3) < PreRel) <= (Nx+deg3)
    Mbj3(i,:) = Z03(1:deg3+1)';

end %end of the loop for i=1:NMC

% calculation of mean values of bj coefficients
Bw1 = mean(Mbj1);
Bw2 = mean(Mbj2);
Bw3 = mean(Mbj3);

% correction of raw variances resulting from MC simulations
varBw1 = fval1/(Nx*Ny-deg1-1).*cov(Mbj1);
varBw2 = fval2/(Nx*Ny-deg2-1).*cov(Mbj2);
varBw3 = fval3/(Nx*Ny-deg3-1).*cov(Mbj3);

% calculation of final residuals by means of OLS with initial experimental
% values all_xi and all_yi_init
% order 1
e1 = all_yi_init-(Bw1(1)+Bw1(2).*all_xi);
cmr1 = e1'*e1;
if Nx*Ny > 2
    sr1 = cmr1/(Nx*Ny-deg1-1);
else
    sr1 = 0;
end

% order 2    
e2 = all_yi_init-(Bw2(1)+Bw2(2).*all_xi+Bw2(3).*all_xi.^2);
cmr2 = e2'*e2;
if Nx*Ny > 3
    sr2 = cmr2/(Nx*Ny-deg2-1);
else
    sr2 = 0;
end

% order 3
e3 = all_yi_init-(Bw3(1)+Bw3(2).*all_xi+Bw3(3).*all_xi.^2+Bw3(4).*all_xi.^3);
cmr3 = e3'*e3;
if Nx*Ny > 4
    sr3 = cmr3/(Nx*Ny-deg3-1);
else
    sr3 = 0;
end

% Snedecor variance test to determine the optimal order dopt
if Nx*Ny == 2
    degopt = 1;
end

if Nx*Ny == 3
    F1calc = (Nx*Ny-3)*(cmr1-cmr2)/cmr2;
    F1theo = finv(0.95,1,Nx*Ny-3);
    if F1calc < F1theo
        degopt = 1;
    else
        degopt = 2;
    end
end

if Nx*Ny >= 4
    F1calc = (Nx*Ny-3)*(cmr1-cmr2)/cmr2;
    F2calc = (Nx*Ny-4)*(cmr2-cmr3)/cmr3;
    F1theo = finv(0.95,1,Nx*Ny-3);
    F2theo = finv(0.95,1,Nx*Ny-4);
    if F1calc < F1theo
        degopt = 1;
    else
        if F2calc < F2theo
            degopt = 2;
        else degopt = 3;
        end
    end
end

% adaptation of dopt to account for the shape of the function:
% functions with 1) d = 3 and changes of concavity or 2) d = 2 and change
% of slope are not selected. This is why this test checks changes of sign
% of the derivatives. 

% calculation of 1st and 2nd derivatives
yd2 = Bw2(2) + 2.*Bw2(3).*rangexi;
yd3 = Bw3(2) + 2.*Bw3(3).*rangexi + 3.*Bw3(4).*rangexi.^2;
ydd3 = 2.* Bw3(3) + 6.*Bw3(4).*rangexi;
% if the slope of function with d = 3 shows changes of sign
% or if absolute value of the second derivative is higher than 1e-2,
% then dopt is changed from d = 3 to d = 2
if degopt ==3 && (sum(yd3>0) < length(yd3) || abs(max(ydd3)) > 1e-2)
    degopt = 2;
end
% if the slope of the function with d = 2 is not strictly positive
% then dopt is chnaged from d = 2 to d = 1
if degopt == 2 && sum(yd2>0) < length(yd2)
        degopt = 1;
end

% final results
Y = [degopt;
    1; Bw1(1); Bw1(2); sqrt(varBw1(1,1)); sqrt(varBw1(2,2)); varBw1(1,2); sr1;...
    2; Bw2(1); Bw2(2); Bw2(3); sqrt(varBw2(1,1)); sqrt(varBw2(2,2)); sqrt(varBw2(3,3)); varBw2(1,2); varBw2(1,3); varBw2(2,3); sr2;...
    3; Bw3(1); Bw3(2); Bw3(3); Bw3(4); sqrt(varBw3(1,1)); sqrt(varBw3(2,2)); sqrt(varBw3(3,3)); sqrt(varBw3(4,4));...
    varBw3(1,2); varBw3(1,3); varBw3(1,4); varBw3(2,3); varBw3(2,4); varBw3(3,4); sr3; PreRel; NMC];

% calculation of final output graphs
out1 = polyval(fliplr(Bw1),rangexi);
out2 = polyval(fliplr(Bw2),rangexi);
out3 = polyval(fliplr(Bw3),rangexi);
out1Wpi = 2.*sqrt(varBw1(1,1) + varBw1(2,2).*rangexi.^2+2.*varBw1(1,2).*rangexi + sr1); 

% graph order 1
titre(1) = {'MC 1st order WLS regression'};
titre(2) = cellstr(['y = ' num2str(Bw1(1)) ' + ' num2str(Bw1(2)) ' x']);
% OLS curve
y0LS1 = polyval(fliplr(bj01_init),rangexi);
% WLS curve
yWLS1 = polyval(fliplr(Bw1),rangexi);

createfigureWLS(A(:,3),all_yi_init,rangexi,y0LS1,rangexi,yWLS1,csvfilename,titre);

% graph order 2
titre(1) = {'MC 2nd order WLS regression'};
titre(2) = cellstr(['y = ' num2str(Bw2(1)) ' + ' num2str(Bw2(2)) ' x + '...
    num2str(Bw2(3)) ' x^2']);
% OLS curve
y0LS2 = polyval(fliplr(bj02_init),rangexi);
% WLS curve
yWLS2 = polyval(fliplr(Bw2),rangexi);
createfigureRW(A(:,3),all_yi_init,rangexi,y0LS2,rangexi,yWLS2,csvfilename,titre);

% graph order 3
titre(1) = {'MC 3rd order WLS regression'};
titre(2) = cellstr(['y = ' num2str(Bw3(1)) ' + ' num2str(Bw3(2)) ' x + '...
    num2str(Bw3(3)) ' x^2 + ' num2str(Bw3(4)) ' x^3']);
% OLS curve
y0LS3 = polyval(fliplr(bj03_init),rangexi);
% WLS curve
yWLS3 = polyval(fliplr(Bw3),rangexi);
createfigureRW(A(:,3),all_yi_init,rangexi,y0LS3,rangexi,yWLS3,csvfilename,titre);

% graph of yik variances
titre(1) = {['data ',csvfilename,'.csv']};
titre(2) = {'variance of the yik values'};
createfigureWLSvar(xi,(si.^2)',csvfilename,titre);

% final results
Y = [degopt;
    1; Bw1(1); Bw1(2); sqrt(varBw1(1,1)); sqrt(varBw1(2,2)); varBw1(1,2); sr1;...
    2; Bw2(1); Bw2(2); Bw2(3); sqrt(varBw2(1,1)); sqrt(varBw2(2,2)); sqrt(varBw2(3,3)); varBw2(1,2); varBw2(1,3); varBw2(2,3); sr2;...
    3; Bw3(1); Bw3(2); Bw3(3); Bw3(4); sqrt(varBw3(1,1)); sqrt(varBw3(2,2)); sqrt(varBw3(3,3)); sqrt(varBw3(4,4));...
    varBw3(1,2); varBw3(1,3); varBw3(1,4); varBw3(2,3); varBw3(2,4); varBw3(3,4); sr3; PreRel; NMC];
% save results in WLS123cal.mat variable
save('WLS123cal.mat','Y');


function y = Willorder1(z)
global Nx Ny deg1 xi wxi all_yik all_wyi;
S1 = Ny*sum(wxi.*((z(deg1+2:Nx+deg1+1)-xi).^2));
S2 = 0;
for i = 1:Nx
    for j = 1+(i-1)*Ny:i*Ny
        S2 = S2 + all_wyi(j)*((z(1) + z(2)*z(deg1+1+i) - all_yik(j))^2);
    end
end
y = S1+S2;


function y = Willorder2(z)
global Nx Ny deg2 xi wxi all_yik all_wyi;
S1 = Ny*sum(wxi.*((z(deg2+2:Nx+deg2+1)-xi).^2));
S2 = 0;
for i = 1:Nx
    for j = 1+(i-1)*Ny:i*Ny
        S2 = S2 + all_wyi(j)*((z(1) + z(2)*z(deg2+1+i) + z(3)*z(deg2+1+i)^2 - all_yik(j))^2);
    end
end
y = S1+S2;


function y = Willorder3(z)
global Nx Ny deg3 xi wxi all_yik all_wyi;
S1 = Ny*sum(wxi.*((z(deg3+2:Nx+deg3+1)-xi).^2));
S2 = 0;
for i = 1:Nx
    for j = 1+(i-1)*Ny:i*Ny
        S2 = S2 + all_wyi(j)*((z(1) + z(2)*z(deg3+1+i) + z(3)*z(deg3+1+i)^2 + z(4)*z(deg3+1+i)^3 - all_yik(j))^2);
    end
end
y = S1+S2;



function createfigureWLS(x1, y1, x2, y2, x3, y3, csvfilename, Title)
%JLBK Auto-generated by MATLAB on 23-Mar-2007 08:38:51

%% Create figure
figure1 = figure;
 
%% Create axes
axes1 = axes(...
  'XGrid','on',...
  'XMinorTick','on',...
  'YGrid','on',...
  'YMinorTick','on',...
  'Parent',figure1);
axis(axes1,[0 1.1*max(x1) 0 1.1*max(y1)]);
title(axes1,Title);
xlabel(axes1,'variable X');
ylabel(axes1,'variable Y');
box(axes1,'on');
hold(axes1,'all');
 
%% Create plot
plot1 = plot(...
  x1,y1,...
  'LineStyle','none',...
  'Color','k',...
  'Marker','*',...
  'Parent',axes1);
 
%% Create mutliple lines using matrix input to plot
plot2 = plot(x2,y2);
set(plot2,...
  'Color',[0.4941 0.4941 0.4941],...
  'LineStyle',':',...
  'Color',[0.7 0.7 0.7],...
  'LineWidth',0.5);

plot3 = plot(x3,y3);
set(plot3,...
  'Color',[0.4941 0.4941 0.4941],...
  'LineStyle','-',...
  'Color','b',...
  'LineWidth',1);
 
%% Create legend
legend1 = legend(axes1,{['data ',csvfilename,'.csv'],'OLS','WLS'},'Location','Best');


function createfigureWLSvar(x1, y1, csvfilename, Title)
%JLBK Auto-generated by MATLAB on 23-Mar-2007 08:38:51

%% Create figure
figure1 = figure;
 
%% Create axes
axes1 = axes(...
  'XGrid','on',...
  'XMinorTick','on',...
  'YGrid','on',...
  'YMinorTick','on',...
  'Parent',figure1);
axis(axes1,[0 1.1*max(x1) 0 1.1*max(y1)]);
title(axes1,Title);
xlabel(axes1,'calibration range');
ylabel(axes1,'variance si ^2');
box(axes1,'on');
hold(axes1,'all');
 
%% Create plot
plot1 = plot(...
  x1,y1,...
  'LineStyle',':',...
  'Color','b',...
  'Marker','o',...
  'LineWidth',0.5,...
  'Parent',axes1);
 
%% Create legend
legend1 = legend(axes1,{'initial variance'},'Location','Best');