function [Y,ech] = uMCM(Z,A,chaine,alpha,MatCor,Mmc,distrib)
% (C) JLBK, 09/2018
% Seven mandatory inputs
% 1 Z: matrix of time varying quantities Zi
% 2 A: matrix of constant quantities Aj
% 3 chaine: string representing the measurement process, based on
%   quantities Zi and Aj
% 4 alpha: level of probability for the coverage interval
% 5 MatCor: matrix of correlation between Zi and Aj quantities
%   described in matrices Z and A
% 6 Mmc: length of the samples (i.e. number of Monte Carlo simulations)
% 7 distrib: matrix with N lines describing the pdf of the N quantities Zi and Aj
% Outputs matrix Y
% Y(:,1): best estimate y of the true value of Y
% Y(:,2): standard uncertainty in Y
% Y(:,3:4): boundaries of the shortest coverage interval with probability level alpha
% The optional output varargout is the matrix containing all the calculated Yi values
% with one column with Mmc values for each line the matrix Z

% function Y = f(Z,A) derived from chaine
function y=fonctionMC(Z,A)
y=eval(chaine);
end

% function CIshort to calculate the shortest coverage interval
function y=CIshort(X,alpha,Mmc)
% sort the vector X    
    Xsort = sort(X);
% set q the integer to scan Xsort
    q=floor(alpha*Mmc + 0.5);
% Create the matrix CI with three columns (respectively low boundary, high boundary and width of the interval) 
    for r=1:Mmc-q
        CI(r,1:3)=[Xsort(r) Xsort(r+q) Xsort(r+q)-Xsort(r)];
    end
% Find the boundaries (in columns 1 and 2) of the shortest interval (in column 3) in CI:
y = CI(find(CI(:,3) == min(CI(:,3))),1:2);
end

% function SampPdf to generate samples from a given copula u and a given
% pdf
function y=SampPdf(u,pdfType,X,uX,index)
switch pdfType
    case 0 % no pdf and thus no sample
        y=repmat(X,Mmc,1);
    case 1 % normal pdf N(m,s)
        y=norminv(u(:,index),X,uX);
    case 2 % uniform pdf U(a,b)
        y=unifinv(u(:,index),X-sqrt(3).*uX,X+sqrt(3).*uX);
    case 3 % triangular pdf Tri(a,mean,b)
        pd=makedist('Triangular',X-sqrt(6).*uX,X,X+sqrt(6).*uX);
        y=icdf(pd,u(:,index));
end
end

% intermediate matrices for calculations
% matrix of quantity Z values 
M1=Z(:,1:2:size(Z,2)-1);
% matrix of standard uncertainties u(Z) values
M2=Z(:,2:2:size(Z,2));
% matrix of quantity A values 
P1=A(:,1:2:size(A,2)-1);
% matrix of standard uncertainties u(A) values
P2=A(:,2:2:size(A,2));


% building the samples of quantities Xi and Aj with their respective pdf
% and the required coefficient of correlation by using copula functions.
% Description of pdf information
% 0 : no uncertainty is accounted for the sample
% 1 : normal pdf N(m,s)
% 2 : uniform pdf U(a,b)
% 3 : triangular pdf Tri(a,b)
% [4,beta] : trapezoidal pdf Trap(a,b,beta)

tic;
% preallocation of Y and varargout
Y=zeros(size(M1,1),4);
ech=zeros(Mmc,size(M1,1));

% Monte Carlo simulations for each line k of the time series of the
% quantities Zi
for k=1:size(M1,1)
    
% create the copula u
    u=copularnd('Gaussian',MatCor,Mmc);
% create the samples vectors SampVect successively for all quantities Zi
% and Aj
    for index=1:size(M1,2) % quantities Zi
        SampVect(:,index) = SampPdf(u,distrib(index,1),M1(k,index),M2(k,index),index);
    end
    for index=1:size(P1,2)
        SampVect(:,index+size(M1,2)) = SampPdf(u,distrib(index+size(M1,2),1),P1(1,index),P2(1,index),index+size(M1,2));
    end
% calculation of values Y
    ech(:,k)=fonctionMC(SampVect(:,1:size(M1,2)),SampVect(:,size(M1,2)+1:size(M1,2)+size(P1,2)));
    Y(k,1:2)=[mean(ech(:,k)) std(ech(:,k))];
% shortest coverage interval
    Y(k,3:4) = CIshort(ech(:,k),alpha,Mmc);
end
toc;
end